""" Module for food intervention models and interfaces with external packages
"""

import xarray as xr
import numpy as np
# from .food_supply import FoodBalanceSheet
import warnings

def balanced_scaling(fbs, items, scale, element, year=None, adoption=None, 
                     timescale=10, origin=None, constant=False,
                     fallback=None):
    """Scale items quantities across multiple elements in a FoodBalanceSheet
    Dataset 
    
    Scales selected item quantities on a food balance sheet and with the
    posibility to keep the sum of selected elements constant.
    Optionally, produce an Dataset with a sequence of quantities over the years
    following a smooth scaling according to the selected functional form.

    The elements used to supply the modified quantities can be selected to keep
    a balanced food balance sheet.

    Parameters
    ----------
    fbs : xarray.Dataset
        Input food balance sheet Dataset.
    items : list
        List of items to scale in the food balance sheet.
    element : string
        Name of the DataArray to scale.
    scale : float
        Scaling parameter after full adoption.
    year : int, optional
        Year of the Food Balance Sheet to use. If not set, the last year of the 
        array is used
    adoption : string, optional
        Shape of the scaling adoption curve. "logistic" uses a logistic model
        for a slow-fast-slow adoption. "linear" uses a constant slope adoption
        during the the "timescale period"
    timescale : int, optional
        Timescale for the scaling to be applied completely.  If "year" +
        "timescale" is greater than the last year in the array, it is extended
        to accomodate the extra years.
    origin : string, optional
        Name of the DataArray which will be used to balance the food balance
        sheets. Any change to the "element" DataArray will be reflected in this
        DataArray.
    constant : bool, optional
        If set to True, the sum of element remains constant by scaling the non
        selected items accordingly.
    fallback : string, optional
        Name of the DataArray used to provide the excess required to balance the
        food balance sheet in case the "origin" falls below zero.

    Returns
    -------
    data : xarray.Dataarray
        Food balance sheet Dataset with scaled "food" values.
    """

    # Check for single item inputs
    if np.isscalar(items):
        items = [items]

    # Check for single item list fbs
    input_item_list = fbs.Item.values
    if np.isscalar(input_item_list):
        input_item_list = [input_item_list]
        if constant:
            warnings.warn("Constant set to true but input only has a single item.")
            constant = False

    # If no items are provided, we scale all of them.
    if items is None or np.sort(items) is np.sort(input_item_list):
        items = fbs.Item.values
        if constant:
            warnings.warn("Cannot keep food constant when scaling all items.")
            constant = False

    # Define Dataarray to use as pivot
    if "Year" in fbs.dims:
        if year is None:
            if np.isscalar(fbs.Year.values):
                year = fbs.Year.values
                fbs_toscale = fbs
            else:
                year = fbs.Year.values[-1]
                fbs_toscale = fbs.isel(Year=-1)
        else:
            fbs_toscale = fbs.sel(Year=year)

    else:
        fbs_toscale = fbs
        try:
            year = fbs.Year.values
        except AttributeError:
            year=0

    # Define scale array based on year range
    if adoption is not None:
        if adoption == "linear":
            from agrifoodpy.utils.scaling import linear_scale as scale_func
        elif adoption == "logistic":
            from agrifoodpy.utils.scaling import logistic_scale as scale_func
        else:
            raise ValueError("Adoption must be one of 'linear' or 'logistic'")
        
        scale_arr = scale_func(year, year, year+timescale-1, year+timescale-1,
                               c_init=1, c_end = scale)
        
        fbs_toscale = fbs_toscale * xr.ones_like(scale_arr)
    
    else:
        scale_arr = scale

    # Create a deep copy to modify and return
    out = fbs_toscale.copy(deep=True)
    osplit = origin.split("-")[-1]
    
    out = out.fbs.scale_add(element, osplit, scale_arr, items, 
                            add = origin.startswith("-"))
    

    if constant:

        delta = out[element] - fbs_toscale[element]

        # Scale non selected items
        non_sel_items = np.setdiff1d(fbs_toscale.Item.values, items)
        non_sel_scale = (fbs_toscale.sel(Item=non_sel_items)[element].sum(dim="Item") - delta.sum(dim="Item")) / fbs_toscale.sel(Item=non_sel_items)[element].sum(dim="Item")
        
        # Make sure inf and nan values are not scaled
        non_sel_scale = non_sel_scale.where(np.isfinite(non_sel_scale)).fillna(1.0)

        if np.any(non_sel_scale < 0):
            warnings.warn("Additional consumption cannot be compensated by \
                        reduction of non-selected items")
        
        out = out.fbs.scale_add(element, osplit, non_sel_scale,
                        non_sel_items, add = origin.startswith("-"))

        # If fallback is defined, adjust to prevent negative values
        if fallback is not None:
            df = out[osplit].where(out[osplit] < 0).fillna(0)
            out[fallback.split("-")[-1]] -= np.where(fallback.startswith("-"), 1, -1)*df
            out[osplit] = out[osplit].where(out[osplit] > 0, 0)

    return out


import numpy as np
import xarray as xr

def test_fbs_impacts():

    from agrifoodpy.impact.model import fbs_impacts

    # Basic test
    items = ["Beef", "Apples"]
    years = [2020, 2021]

    fbs = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[10, 20], [30, 40]]),
            exports=(["Year", "Item"], [[5, 10], [15, 20]]),
            production=(["Year", "Item"], [[50, 60], [70, 80]])
            ),

    coords=dict(Item=("Item", items), Year=("Year", years))
    )

    impact = xr.DataArray([100, 0.5], dims=("Item"), coords={"Item": items})

    result = fbs_impacts(fbs, impact)

    truth = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[1000, 10], [3000, 20]]),
            exports=(["Year", "Item"], [[500, 5], [1500, 10]]),
            production=(["Year", "Item"], [[5000, 30], [7000, 40]])
            ),
        coords=dict(Item=("Item", items), Year=("Year", years))
    )

    assert result.equals(truth)

    # Test with population
    population = xr.DataArray(
        data = [1.0e6, 1.1e6],
        dims= ("Year"),
        coords={"Year": years})

    result = fbs_impacts(fbs, impact, population=population)

    truth = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[1e9, 1e7], [3.3e9, 2.2e7]]),
            exports=(["Year", "Item"], [[5e8, 5e6], [1.65e9, 1.1e7]]),
            production=(["Year", "Item"], [[5e9, 3e7], [7.7e9, 4.4e7]])
            ),
        coords=dict(Item=("Item", items),
                    Year=("Year", years))
        )
        
    assert result.equals(truth)

    # Test summing over dimensions

    sum_dims = ["Item"]

    result = fbs_impacts(fbs, impact, sum_dims=sum_dims)

    truth = xr.Dataset(
        data_vars=dict(
            imports=(["Year"], [1.01e3, 3.02e3]),
            exports=(["Year"], [5.05e2, 1.51e3]),
            production=(["Year"], [5.03e3, 7.04e3])
            ),
        coords=dict(Year=("Year", years))
        )
        
    assert result.equals(truth)

def test_fair_interface():

    from agrifoodpy.impact.model import fair_co2_only

    # Test with single emission

    emissions = xr.DataArray(10, coords={"Year":2015})

    T, C, F = fair_co2_only(emissions=emissions)

    T_truth = xr.DataArray([0.0, 0.00133196], 
                           coords={"timebounds":[2014.5, 2015.5]})
    C_truth = xr.DataArray([278.3, 279.3534663],
                           coords={"timebounds":[2014.5, 2015.5]})
    F_truth = xr.DataArray([0.0, 0.02122413],
                           coords={"timebounds":[2014.5, 2015.5]})

    xr.testing.assert_allclose(T, T_truth)
    xr.testing.assert_allclose(C, C_truth)
    xr.testing.assert_allclose(F, F_truth)

    # Test with multiple emission values

    emissions = xr.DataArray(np.repeat(35, 10),
                             coords={"Year":np.arange(2010,2020)})

    T, C, F = fair_co2_only(emissions=emissions)

    T_truth = xr.DataArray([0.        , 0.00464001, 0.01451917, 0.02775184,
                            0.04310335, 0.05982477, 0.07744928, 0.09567989,
                            0.11432504, 0.13325952, 0.15239999],
                        coords={"timebounds":np.linspace(2009.5, 2019.5, 11)})

    C_truth = xr.DataArray([278.3       , 281.98713205, 284.96432983,
                            287.67924614, 290.23401333, 292.66883373,
                            295.00890336, 297.27301864, 299.47583696,
                            301.62892913, 303.74148527],
                        coords={"timebounds":np.linspace(2009.5, 2019.5, 11)})

    F_truth = xr.DataArray([0.        , 0.07393624, 0.13293446, 0.18620025,
                            0.23586687, 0.28279642, 0.32753321, 0.37048144,
                            0.41195413, 0.45219685, 0.4914037],
                        coords={"timebounds":np.linspace(2009.5, 2019.5, 11)})

    xr.testing.assert_allclose(T, T_truth)
    xr.testing.assert_allclose(C, C_truth)
    xr.testing.assert_allclose(F, F_truth)



import numpy as np
import xarray as xr
from agrifoodpy.land.land import LandDataArray
import pytest

def test_area_by_type():
    
    data = np.tile((np.arange(3, dtype=float)), (4,1))

    da = xr.DataArray(data=data, name="land",
                    coords={"x": [0, 1, 2, 3], "y": [0, 1, 2]})

    land = LandDataArray(da)
    # Test area by type without explicit values
    expected_result_no_values = np.array([4,4,4])
    result_no_values = land.area_by_type()
    
    assert(np.array_equal(expected_result_no_values, result_no_values))
    assert(list(result_no_values.coords) == ["land"])

    # Test area by type with explicit values
    expected_result_values = np.array([4,4])
    result_values = land.area_by_type(values=[1,2])
    
    assert(np.array_equal(expected_result_values, result_values))
    assert(list(result_values.coords) == ["land"])

    # Test area by type with explicit dimension name
    dim_name = "area"
    expected_result_dims = np.array([4,4,4])
    result_dims = land.area_by_type(dim=dim_name)
    
    assert(np.array_equal(expected_result_dims, result_dims))
    assert(list(result_dims.coords) == [dim_name])

    # Test area by type with nan values
    da_nan = da.copy(deep=True)
    da_nan[-1, -1] = np.nan
    da_nan.name = "land_nan"
    land_nan = LandDataArray(da_nan)

    expected_result_nan = np.array([4,4,3])
    result_nan = land_nan.area_by_type()

    assert(np.array_equal(expected_result_nan, result_nan))
    assert(list(result_nan.coords) == ["land_nan"])

def test_area_overlap():

    data_left = np.tile((np.arange(3, dtype=float)), (4,1))
    data_right = np.repeat((np.arange(4, dtype=float)), 3).reshape((4,3))

    name_left, name_right = "land_left", "land_right"

    da_left = xr.DataArray(data=data_left, name=name_left,
                    coords={"x": [0, 1, 2, 3], "y": [0, 1, 2]})
    
    da_right = xr.DataArray(data=data_right, name=name_right,
                    coords={"x": [0, 1, 2, 3], "y": [0, 1, 2]})

    land_left = LandDataArray(da_left)
    land_right = LandDataArray(da_right)

    # Test area overlap without explicit values

    expected_results_no_values = np.ones((3,4))
    result_no_values = land_left.area_overlap(da_right)

    assert(np.array_equal(expected_results_no_values, result_no_values))
    assert(list(result_no_values.coords) == [name_left, name_right])

    # Test area overlap with explicit values
    input_values_left = [1,2]
    input_values_right = [2,3]

    expected_results_values = np.ones((2,2))
    result_values = land_left.area_overlap(da_right,
                                           values_left=input_values_left,
                                           values_right=input_values_right)
    
    assert(np.array_equal(expected_results_no_values, result_no_values))
    assert(np.array_equal(result_no_values.coords, [name_left, name_right]))
    assert(np.array_equal(result_values[name_left].values, input_values_left))
    assert(np.array_equal(result_values[name_right].values, input_values_right))

    # Test area overlap with explicit dimension names

    dim_left, dim_right = "category_left", "category_right"

    expected_results_dimensions = np.ones((3,4))
    result_dimensions = land_left.area_overlap(da_right,
                                              dim_left=dim_left,
                                              dim_right=dim_right)

    assert(np.array_equal(expected_results_dimensions, result_dimensions))
    assert(np.array_equal(result_dimensions.coords, [dim_left, dim_right]))

    # Test area overlap with nan values

    da_left_nan = da_left.copy(deep=True)
    da_left_nan[-1,-1] = np.nan
    da_left_nan.name = "land_left_nan"
    land_left_nan = LandDataArray(da_left_nan)

    da_right_nan = da_right.copy(deep=True)
    da_right_nan[0,0] = np.nan
    da_right_nan.name = "land_right_nan"

    expected_results_nan = np.ones((3,4))
    expected_results_nan[-1,-1] = 0
    expected_results_nan[0,0] = 0

    result_nan = land_left_nan.area_overlap(da_right_nan)

    assert(np.array_equal(expected_results_nan, result_nan))
    
def test_category_match():

    data_left = np.tile((np.arange(3, dtype=float)), (4,1))
    data_right = np.repeat((np.arange(4, dtype=float)), 3).reshape((4,3))

    name_left, name_right = "land_left", "land_right"

    da_left = xr.DataArray(data=data_left, name=name_left,
                    coords={"x": [0, 1, 2, 3], "y": [0, 1, 2]})
    
    da_right = xr.DataArray(data=data_right, name=name_right,
                    coords={"x": [0, 1, 2, 3], "y": [0, 1, 2]})

    land_left = LandDataArray(da_left)
    land_right = LandDataArray(da_right)

    # Basic example without set values
    result_basic = land_left.category_match(da_right)
    assert result_basic.equals(da_left)
    assert result_basic.name == name_left

    # Example with values on left map
    result_values_left = land_left.category_match(da_right, values_left=1)
    assert result_values_left.where(result_values_left==1).equals(da_left.where(da_left==1))
    assert np.all(result_values_left.where(result_values_left!=1).isnull())

    # Example with values on right map
    result_values_left = land_left.category_match(da_right, values_right=1)
    
    # Example with multiple values on both maps
    result_multivar = land_left.category_match(da_right,
                                               values_left=[0,1],
                                               values_right=[1,2])

    non_nan_values = result_multivar.where(~np.isnan(result_multivar),
                                           drop=True)
    truth = np.array([[0., 1.], [0., 1.]])
    assert np.array_equal(non_nan_values, truth)
    assert np.array_equal(non_nan_values.x.values, [1, 2])
    assert np.array_equal(non_nan_values.y.values, [0, 1])
    assert result_multivar.name == name_left

    # Example with non-matching values
    result_non_matching = land_left.category_match(da_right, values_left=4)
    assert np.all(result_non_matching.isnull())



""" Food supply module.

The Food module provides the FoodBalanceSheet and FoodElementSheet accessor
classes to manipulate and analyse Food data stored in xarray.Dataset and 
xarray.DataArray formats, respectively.

It also provides a constructor style function which allows the creation of a
FoodBalanceSheet or FoodElementSheet style xarray primitive.
"""

import numpy as np
import xarray as xr
import copy

from agrifoodpy.array_accessor import XarrayAccessorBase

import matplotlib.pyplot as plt

def FoodSupply(items, years, quantities, regions=None, elements=None,
               long_format=True):
    """ Food Supply style dataset constructor

    Constructs a food balance sheet style xarray.Dataset or xarray.DataArray for
    a given list of items, years and regions, and an array data shaped
    accordingly.  
     
    Parameters
    ----------
    items : (ni,) array_like
        The Item identifying ID or strings for which coordinates will be
        created.
    years : years : (ny,) array_like
        The year values for which coordinates will be created.
    quantities : ([ne], ni, ny, [nr]) ndarray
        Array containing the food quantity for each combination of `Item`,
        `Year` and, optionally `Region`.
    regions : (nr,) array_like, optional
        The region identifying ID or strings for which coordinates will be
        created.
    long_format : bool
        Boolean flag to interpret data in long or wide format
    elements : (ne,) array_like, optional
        Array with element name strings. If `elements` is provided, a dataset
        is created for each element in `elements` with the quantities being each
        of the sub-arrays indexed by the first coordinate of the input array.

    Returns
    -------
    fbs : xarray.Dataset
        Food Supply dataset containing the food quantity for each `Item`, `Year`
        and `Region` with one dataarray per element in `elements`.
    """

    # if the input has a single element, proceed with long format
    if np.isscalar(quantities):
        long_format = True

    quantities = np.array(quantities)

    # Identify unique values in coordinates
    _items = np.unique(items)
    _years = np.unique(years)
    coords = {"Item" : _items,
              "Year" : _years,} 

    # find positions in output array to organize data
    ii = [np.searchsorted(_items, items), np.searchsorted(_years, years)]
    size = (len(_items), len(_years))

    # If regions and are provided, add the coordinate information 
    if regions is not None:
        _regions = np.unique(regions)
        ii.append(np.searchsorted(_regions, regions))
        size = size + (len(_regions),)
        coords["Region"] = _regions

    # Create empty dataset
    fbs = xr.Dataset(coords = coords)

    if long_format:
        # dataset, quantities
        ndims = 2
    else:
        # dataset, coords
        ndims = len(coords)+1
    
    # make sure the long format has the right number of dimensions
    while len(quantities.shape) < ndims:
        quantities = np.expand_dims(quantities, axis=0)

    # If no elements names are given, then create generic ones,
    # one for each dataset
    if elements is None:
        elements = [f"Quantity {id}" for id in range(quantities.shape[0])]

    # Else, if a single string is given, transform to list. If doesn't match
    # the number of datasets created above, xarray will return an error.
    elif isinstance(elements, str):
        elements = [elements]

    if long_format:
        # Create a datasets, one at a time
        for ie, element in enumerate(elements):
            values = np.zeros(size)*np.nan
            values[tuple(ii)] = quantities[ie]
            fbs[element] = (coords, values)

    else:
        quantities = quantities[:, ii[0]]
        if len(quantities.shape) < ndims:
            quantities = np.expand_dims(quantities, axis=0)

        quantities = quantities[:, :, ii[1]]
        if len(quantities.shape) < ndims:
            quantities = np.expand_dims(quantities, axis=0)

        if regions is not None:
            quantities = quantities[..., ii[2]]
        # Create a datasets, one at a time
        for ie, element in enumerate(elements):
            values = quantities[ie]
            fbs[element] = (coords, values)

    return fbs

@xr.register_dataset_accessor("fbs")
class FoodBalanceSheet(XarrayAccessorBase):

    def scale_element(self, element, scale, items=None):
        """Scales list of items from an element in a food balance sheet like
        DataSet.

        Parameters
        ----------
        fbs : xarray.Dataset
            Input DataSet with FAOSTAT like elements
        scale : float, float array_like or xarray.Dataarray
            Scaling quantities for the element DataArray
        element_in : str
            Element DataArray to be scaled
        element_out : str
            Destination element DataArray to which the difference is added to
        items : list of int or list of str, optional
            List of items to be scaled. If not provided, all items are scaled.
            
        Returns
        -------
        out : xarray.Dataset
            FAOSTAT formatted Food Supply dataset with scaled quantities.
        
        """

        fbs = self._obj

        if np.isscalar(items):
            items = [items]

        input_item_list = fbs.Item.values.tolist()

        # If no item list provided, or list is all the items of the fbs
        if items is None or np.sort(items) is np.sort(input_item_list):
            items = fbs.Item.values

        # Create copy to return
        out = copy.deepcopy(fbs)

        # Scale items
        sel = {"Item":items}

        out[element].loc[sel] = out[element].loc[sel] * scale

        return out

    def scale_add(self, element_in, element_out, scale, items=None, add=True):
        """Scales item quantities of an element and adds the difference to
        another element DataArray
        
        Parameters
        ----------
        fbs : xarray.Dataset
            Input DataSet with FAOSTAT like elements
        element_in : str
            Element DataArray to be scaled
        element_out : str
            Destination element DataArray to which the difference is added to
        scale : float, float array_like or xarray.Dataarray
            Scaling quantities for the element DataArray
        items : list of int or list of str, optional
            List of items to be scaled. If not provided, all items are scaled.
        add : boolean
            Wether to add or subtract the difference to element_out 
            
        Returns
        -------
        out : xarray.Dataset
            FAOSTAT formatted Food Supply dataset with scaled quantities.

        """

        fbs = self._obj

        out = self.scale_element(element_in, scale, items)
        dif = fbs[element_in].fillna(0) - out[element_in].fillna(0)

        out[element_out] += np.where(add, -1, 1)*dif
        
        return out

    def SSR(self, items=None, per_item=False, production="production",
            imports="imports", exports="exports"):
        """Self-sufficiency ratio

        Self-sufficiency ratio (SSR) or ratios for a list of item imports,
        exports and production quantities.

        Parameters
        ----------
        fbs : xarray.Dataset
            Input Dataset containing an "Item" coordinate and, optionally, a
            "Year" coordinate.
        items : list, optional
            list of items to compute the SSR for from the food Dataset. If no
            list is provided, the SSR is computed for all items.
        per_item : bool, optional
            Whether to return an SSR for each item separately. Default is false
        production : string, optional
            Name of the DataArray containing the production data
        imports : string, optional
            Name of the DataArray containing the imports data
        exports : string, optional
            Name of the DataArray containing the exports data

        Returns
        -------
        data : xarray.Dataarray
            Self-sufficiency ratio or ratios for the list of items, one for each
            year of the input food Dataset "Year" coordinate.

        """

        fbs = self._obj

        if items is not None:
            if np.isscalar(items):
                items = [items]
            fbs = fbs.sel(Item=items)

        if per_item:
            return fbs[production] / (fbs[production] + fbs[imports] - fbs[exports])

        return fbs[production].sum(dim="Item") / (fbs[production]+fbs[imports]-fbs[exports]).sum(dim="Item")

    def IDR(self, items=None, per_item=False):
        """Import-dependency ratio

        Import-ependency ratio (IDR) or ratios for a list of item imports,
        exports and production quantities.

        Parameters
        ----------
        fbs : xarray.Dataset
            Input Dataset containing an "Item" coordinate and, optionally, a
            "Year" coordinate.
        items : list, optional
            list of items to compute the IDR for from the food Dataset. If no
            list is provided, the IDR is computed for all items.
        per_item : bool, optional
            Whether to return an IDR for each item separately. Default is false.

        Returns
        -------
        data : xarray.Datarray
            Import-dependency ratio or ratios for the list of items, one for
            each year of the input food Dataset "Year" coordinate.

        """

        fbs = self._obj

        if items is not None:
            if np.isscalar(items):
                items = [items]
            fbs = fbs.sel(Item=items)

        if per_item:
            return fbs["imports"] / (fbs["production"] + fbs["imports"] - fbs["exports"])

        return fbs["imports"].sum(dim="Item") / (fbs["production"]+fbs["imports"]-fbs["exports"]).sum(dim="Item")

    def plot_bars(self, show="Item", elements=None, inverted_elements=None,
                  ax=None, colors=None, labels=None, **kwargs):
        """Plot total quantities per element on a horizontal bar plot

        Produces a horizontal bar plot with a bar per element on the vertical
        axis plotted on a cumulative form. Each bar is the sum of quantities on
        each element, broken down by the selected coordinate "show". The
        starting x-axis position of each bar will depend on the cumulative value
        up to that element. The order of elements can be defined by the
        "element" parameter. A second set of "inverted_elements" can be given,
        and these will be plotted from right to left starting from the previous
        cumulative sum, minus the corresponding sum of the inverted elements. 

        Parameters
        ----------
        show : str, optional
            Name of the coordinate to dissagregate when filling the horizontal
            bar. The quantities are summed along the remaining coordinates.
        elements : str list, optional
            List of DataArray names in the Dataset to plot in ascending
            cumulative sum from left to right and top to bottom. If not
            provided, all DataArrays are plotted.
        inverted_elements : strr list, optional
            List of DataArray names in the Dataset to plot in descending
            cumulative sum from right to left, and top to bottom. If not
            provided, none of the DataArray is used.
        ax : matplotlib.pyplot.artist, optional
            Axes on which to draw the plot. If not provided, a new artist is
            created.
        colors : list of str, optional
            String list containing the colors for each of the elements in the
            "show" coordinate.
            If not defined, a color list is generated from the standard cycling.
        labels : str, list of str, optional
            String list containing the labels for the legend of the elements in
            the "show" coordinate. If not set, no labels are printed. If "show",
            the values of the "show" dimension are used.
        **kwargs : dict
            Style options to be passed on to the actual plot function, such as
            linewidth, alpha, etc.

        Returns
        -------
            ax : matplotlib axes instance
        """

        fbs = self._obj

        if ax is None:
            f, ax = plt.subplots(**kwargs)

        if elements is None:
            elements = list(fbs.keys())
            plot_elements = elements
        elif np.isscalar(elements):
            elements = [elements]

        len_elements = len(elements)

        # Define dimensions to sum over
        bar_dims = list(fbs.dims)
        if show in bar_dims:
            bar_dims.remove(show)
            size_show = fbs.sizes[show]
        else:
            size_show = 1

        # Make sure NaN and inf do not interfere
        fbs = fbs.fillna(0)
        fbs = fbs.where(np.isfinite(fbs), other=0)

        food_sum = fbs.sum(dim=bar_dims)

        # If colors are not defined, generate a list from the standard cycling
        if colors is None:
            colors = [f"C{ic}" for ic in range(size_show)]

        # If labels are not defined, generate a list from the dimensions
        print_labels = True
        if labels is None:
            # empty label placeholder and no-label flag
            labels = np.empty(len(fbs[show].values), dtype=str)
            print_labels = False

        elif np.all(labels == "show"):
            labels = [str(val) for val in fbs[show].values]


        # Plot non inverted elements first
        cumul = 0
        for ie, element in enumerate(elements):
            ax.hlines(ie, 0, cumul, color="k", alpha=0.2, linestyle="dashed",
                    linewidth=0.5)
            if size_show == 1:
                ax.barh(ie, left = cumul, width=food_sum[element], 
                        color=colors[0])
                cumul +=food_sum[element]
            else:
                for ii, val in enumerate(food_sum[element]):
                    ax.barh(ie, left = cumul, width=val, color=colors[ii],
                            label=labels[ii])
                    cumul += val

        # Then the inverted elements
        if inverted_elements is not None:
            
            if np.isscalar(inverted_elements):
                inverted_elements = [inverted_elements]

            len_elements += len(inverted_elements)
            plot_elements = np.concatenate([elements, inverted_elements])

            cumul = 0
            for ie, element in enumerate(reversed(inverted_elements)):
                ax.hlines(len_elements-1 - ie, 0, cumul, color="k", alpha=0.2,
                        linestyle="dashed", linewidth=0.5)
                if size_show == 1:
                    ax.barh(len_elements-1 - ie, left = cumul,
                            width=food_sum[element], color=colors[0])
                    cumul +=food_sum[element]
                else:
                    for ii, val in enumerate(food_sum[element]):
                        ax.barh(len_elements-1 - ie, left = cumul, width=val,
                                color=colors[ii], label=labels[ii])
                        cumul += val

        # Plot decorations
        ax.set_yticks(np.arange(len_elements), labels=plot_elements)
        ax.tick_params(axis="x",direction="in", pad=-12)
        ax.invert_yaxis()  # labels read top-to-bottom
        ax.set_ylim(len_elements,-1)

        # Unique labels
        if print_labels:
            handles, labels = ax.get_legend_handles_labels()
            by_label = dict(zip(labels, handles))
            ax.legend(by_label.values(), by_label.keys(), fontsize=6)

        return ax

@xr.register_dataarray_accessor("fbs")
class FoodElementSheet(XarrayAccessorBase):
        
    def plot_years(self, show=None, stack=True,ax=None, colors=None,
                   labels=None, **kwargs):
        """ Fill plot with quantities at each year value

        Produces a vertical fill plot with quantities for each year on the
        "Year" coordinate of the input dataset in the horizontal axis. If the
        "show" coordinate exists, then the vertical fill plot is a stack of the 
        sums of the other coordinates at that year for each item in the "show"
        coordinate.

        Parameters
        ----------
        food : xarray.Dataarray
            Input Dataarray containing a "Year" coordinate and optionally,
            additional coordinates to dissagregate.
        show : str, optional
            Name of the coordinate to dissagregate when filling the vertical
            plot. The quantities are summed along the remaining coordinates.
            If the coordinate is not provided or does not exist in the input,
            all coordinates are summed and a plot with a single fill curve is
            returned.
        stack : boolean, optional
            Whether to stack fill plots or not. If 'True', the fill curves are
            stacked on top of each other and the upper fill curve represents the
            sum of all elements for a given year.
            If 'false', each element along the 'show' dimension is plotted
            starting from the origin.         
        ax : matplotlib.pyplot.artist, optional
            Axes on which to draw the plot. If not provided, a new artist is
            created.
        colors : list of str, optional
            String list containing the colors for each of the elements in the
            "show" coordinate.
            If not defined, a color list is generated from the standard cycling.
        labels : str, list of str, optional
            String list containing the labels for the legend of the elements in
            the "show" coordinate. If not set, no labels are printed. If "show",
            the values of the "show" dimension are used.
        **kwargs : dict
            Style options to be passed on to the actual plot function, such as
            linewidth, alpha, etc.

        Returns
        -------
            ax : matplotlib axes instance
        """

        fbs = self._obj

        # If no years are found in the dimensions, raise an exception
        sum_dims = list(fbs.dims)
        if "Year" not in sum_dims:
            raise TypeError("'Year' dimension not found in array data")

        # Define new ax if not provided
        if ax is None:
            f, ax = plt.subplots(1, **kwargs)

        # Define the cumsum and sum dims and check for one element dims
        sum_dims.remove("Year")
        if show in sum_dims:
            sum_dims.remove(show)
            size_cumsum = fbs.sizes[show]
            if stack:
                cumsum = fbs.cumsum(dim=show).transpose(show, ...)
            else:
                cumsum = fbs
        else:
            size_cumsum = 1
            cumsum = fbs

        # Collapse remaining dimensions
        cumsum = cumsum.sum(dim=sum_dims)
        years = fbs.Year.values

        # If colors are not defined, generate a list from the standard cycling
        if colors is None:
            colors = [f"C{ic}" for ic in range(size_cumsum)]

        # If labels are not defined, generate a list from the dimensions
        print_labels = True
        if labels is None:
            # empty label placeholder and no-label flag
            if show is not None:
                labels = np.empty(len(fbs[show].values))
            else:
                labels = np.empty(len(fbs["Year"].values))
            print_labels = False

        elif np.all(labels == "show"):
            labels = fbs[show].values

        # Plot
        if size_cumsum == 1:
            ax.fill_between(years, cumsum, color=colors[0], alpha=0.5)
            ax.plot(years, cumsum, color=colors[0], linewidth=0.5, label=labels)
        else:
            for id in reversed(range(size_cumsum)):
                ax.fill_between(years, cumsum.isel({show:id}), color=colors[id], alpha=0.5)
                ax.plot(years, cumsum.isel({show:id}), color=colors[id], linewidth=0.5,
                        label=labels[id])

        ax.set_xlim(years.min(), years.max())
        ax.set_ylim(bottom=0)
        ax.set_xlabel("Year")

        if print_labels:
            ax.legend()

        return ax


import numpy as np
import xarray as xr
from agrifoodpy.food.food import FoodBalanceSheet
from agrifoodpy.food.food import FoodSupply
import pytest

def test_FoodSupply():

    # Single item, single year
    single_item = "chicken"
    single_year = 1990
    single_quantity = 10

    result = FoodSupply(items=single_item, years=single_year,
                        quantities=single_quantity)

    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year"],
                                                  [[single_quantity]])},
                    coords = {"Year":[single_year], "Item":[single_item]})

    assert truth.equals(result)

    # Single array item, single array year
    single_array_item = ["chicken"]
    single_array_year = [1990]
    single_array_quantity = [10]

    result = FoodSupply(items=single_array_item, years=single_array_year,
                        quantities=single_array_quantity)

    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year"],
                                                  [single_array_quantity])},
                    coords = {"Year":single_array_year,
                              "Item":single_array_item})

    assert truth.equals(result)

    # Single item, single year, single region
    single_region = "UK"
    result = FoodSupply(items=single_item, years=single_year,
                        regions=single_region, quantities=single_quantity)

    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year", "Region"],
                                                  [[[single_quantity]]])},
                    coords = {"Item":[single_item],
                              "Year":[single_year],
                              "Region":[single_region]})

    assert truth.equals(result)

    # Single element dimensions, named element 
    single_element = "Food"

    result = FoodSupply(items=single_item, years=single_year,
                        regions=single_region, quantities=single_quantity,
                        elements=single_element)

    truth = xr.Dataset(data_vars = {single_element:(["Item", "Year", "Region"],
                                                    [[[single_quantity]]])},
                    coords = {"Item":[single_item],
                              "Year":[single_year],
                              "Region":[single_region]})

    assert truth.equals(result)

    # Single item, many years, single region
    many_years = [1990, 1991, 1992]
    many_years_qty = [10,20,30]

    result = FoodSupply(items=single_item, years=many_years,
                        regions=single_region, quantities=many_years_qty)
    
    truth_array = np.array(many_years_qty).reshape(1,3,1)
    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year", "Region"],
                                                  truth_array)},
                       coords = {"Item":[single_item],
                                 "Year":many_years,
                                 "Region":[single_region]},)

    assert truth.equals(result)

    # Multiple elements in all dimensions
    many_dim_years = np.array([1990, 1991, 1992, 1992])
    many_dim_items = ["chicken", "chicken", "chicken", "beef"]
    many_dim_regions = ["UK", "UK", "UK", "USA"]
    many_dims_qty = [10, 20, 30, 40]

    result = FoodSupply(items=many_dim_items, years=many_dim_years,
                        regions=many_dim_regions, quantities=many_dims_qty)
    
    truth_array = np.array([[[np.nan,np.nan],[np.nan,np.nan],[np.nan,40]],
                        [[10,np.nan],[20,np.nan],[30,np.nan]]])

    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year", "Region"],
                                                  truth_array)},
                       coords = {"Item":np.unique(many_dim_items),
                                 "Year":np.unique(many_dim_years),
                                 "Region":np.unique(many_dim_regions)})

    assert truth.equals(result)

    # TODO Multiple element test. Have to fix the code to be truly long format
    # friendly

    # Many elements in all dimensions, wide format

    many_dim_years = np.array([1990, 1991])
    many_dim_items = ["beef", "chicken"]
    many_dim_regions = ["UK", "USA"]
    many_dim_qty = 10*np.arange(8).reshape(2,2,2)

    truth = xr.Dataset(data_vars = {"Quantity 0":(["Item", "Year", "Region"],
                                                  many_dim_qty)},
                    coords = {"Item":many_dim_items,
                              "Year":many_dim_years,
                              "Region":many_dim_regions})

    result = FoodSupply(items=many_dim_items, years=many_dim_years,
                        regions=many_dim_regions, quantities=many_dim_qty,
                        long_format=False)

    assert result.equals(truth)

    # Many elements in all dimensions, multiple named elements, wide format

    many_dim_years = np.array([1990, 1991])
    many_dim_items = ["beef", "chicken"]
    many_dim_regions = ["UK", "USA"]
    many_dim_qty = 10*np.arange(16).reshape(2,2,2,2)
    many_dim_elements = ["Food", "Waste"]

    result = FoodSupply(items=many_dim_items, years=many_dim_years,
                        regions=many_dim_regions, quantities=many_dim_qty,
                        elements=many_dim_elements, long_format=False)

    truth = xr.Dataset(data_vars = {"Food":(["Item", "Year", "Region"],
                                            many_dim_qty[0]),
                                    "Waste": (["Item", "Year", "Region"],
                                              many_dim_qty[1])},
                    coords = {"Item":many_dim_items,
                              "Year":many_dim_years,
                              "Region":many_dim_regions})

    assert result.equals(truth)

def test_SSR():

    items = ["Beef", "Apples"]
    years = [2020, 2021]

    ds = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[10, 20], [30, 40]]),
            exports=(["Year", "Item"], [[5, 10], [15, 20]]),
            production=(["Year", "Item"], [[50, 60], [70, 80]])
            ),

    coords=dict(Item=("Item", items), Year=("Year", years))
    )

    fbs = FoodBalanceSheet(ds)

    # Test basic result on all items

    result_basic = fbs.SSR()
    ex_result_basic = xr.DataArray([0.88, 0.810810], dims=("Year"),
                                coords={"Year": years})

    xr.testing.assert_allclose(result_basic, ex_result_basic)

    # Test for an item subset
    result_subset = fbs.SSR(items="Beef")
    ex_result_subset = xr.DataArray([0.909090, 0.823529], dims=("Year"),
                                    coords={"Year": years})

    xr.testing.assert_allclose(result_subset, ex_result_subset)

    # Test per item
    result_peritem = fbs.SSR(per_item=True)
    ex_result_peritem = xr.DataArray([[0.909090, 0.857142], [0.823529, 0.8]],
                                    dims=(["Year", "Item"]),
                                    coords={"Year": years, "Item": items})

    xr.testing.assert_allclose(result_peritem, ex_result_peritem)

def test_IDR():

    items = ["Beef", "Apples"]
    years = [2020, 2021]

    ds = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[10, 20], [30, 40]]),
            exports=(["Year", "Item"], [[5, 10], [15, 20]]),
            production=(["Year", "Item"], [[50, 60], [70, 80]])
            ),
      
    coords=dict(Item=("Item", items), Year=("Year", years))
    )

    fbs = FoodBalanceSheet(ds)

    # Test basic result on all items
    result_basic = fbs.IDR()
    ex_result_basic = xr.DataArray([0.24, 0.37837838], dims=("Year"),
                                   coords={"Year": years})
    
    xr.testing.assert_allclose(result_basic, ex_result_basic)

    # Test for an item subset
    result_subset = fbs.IDR(items="Beef")
    ex_result_subset = xr.DataArray([0.1818181, 0.352941], dims=("Year"),
                                    coords={"Year": years})
    
    xr.testing.assert_allclose(result_subset, ex_result_subset)

    # Test per item
    result_peritem = fbs.IDR(per_item=True)
    ex_result_peritem = xr.DataArray([[0.1818181, 0.285714], [0.352941, 0.4]],
                                     dims=(["Year", "Item"]),
                                     coords={"Year": years, "Item": items})
    
    xr.testing.assert_allclose(result_peritem, ex_result_peritem)

def test_scale_add():

    items = ["Beef", "Apples"]
    years = [2020, 2021]

    scalar_scale = 1.5
    array_scale = np.arange(4).reshape((2,2))

    xarray_year_scale = xr.DataArray([1, 1.5], coords={'Year': years},
                                     dims=['Year'])
    xarray_item_scale = xr.DataArray([1, 1.5], coords={'Item': items},
                                     dims=['Item'])

    elin = "production"
    elout = "imports"

    im_data = np.array([[10, 20], [30, 40]])
    ex_data = np.array([[5, 10], [15, 20]])
    pr_data= np.array([[50, 60], [70, 80]])

    ds = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], im_data),
            exports=(["Year", "Item"], ex_data),
            production=(["Year", "Item"], pr_data)
            ),

    coords=dict(Item=("Item", items), Year=("Year", years))
    )

    fbs = FoodBalanceSheet(ds)

    fbs._obj

    # Test with float scale on all items
    result_scalar = fbs.scale_add(elin, elout, scalar_scale)
    assert np.array_equal(result_scalar[elin], ds[elin]*scalar_scale)
    assert np.array_equal(result_scalar[elout], ds[elout] +
                          (result_scalar[elin] - ds[elin]))

    # Test with float scale on all items with subtraction 
    result_scalar_sub = fbs.scale_add(elin, elout, scalar_scale, add=False)
    assert np.array_equal(result_scalar_sub[elin], ds[elin]*scalar_scale)
    assert np.array_equal(result_scalar_sub[elout], ds[elout] -
                          (result_scalar_sub[elin] - ds[elin]))

    # Test with float scale on some items
    result_scalar_item = fbs.scale_add(elin, elout, scalar_scale, items="Beef")
    assert np.array_equal(result_scalar_item[elin].sel(Item="Beef"),
                          ds[elin].sel(Item="Beef")*scalar_scale)
    
    assert np.array_equal(result_scalar_item[elout].sel(Item="Beef"),
                          ds[elout].sel(Item="Beef") + 
                          (result_scalar_item[elin].sel(Item="Beef")
                           - ds[elin].sel(Item="Beef")))
    
    assert np.array_equal(result_scalar_item[elin].sel(Item="Apples"),
                          ds[elin].sel(Item="Apples"))
    
    assert np.array_equal(result_scalar_item[elout].sel(Item="Apples"),
                          ds[elout].sel(Item="Apples"))

    # Test with float scale on some items with subtraction
    result_scalar_item_sub = fbs.scale_add(elin, elout, scalar_scale,
                                           items="Beef", add=False)
    
    assert np.array_equal(result_scalar_item_sub[elin].sel(Item="Beef"),
                          ds[elin].sel(Item="Beef")*scalar_scale)
    
    assert np.array_equal(result_scalar_item_sub[elout].sel(Item="Beef"),
                          ds[elout].sel(Item="Beef") -
                          (result_scalar_item_sub[elin].sel(Item="Beef")
                           - ds[elin].sel(Item="Beef")))
    
    assert np.array_equal(result_scalar_item_sub[elin].sel(Item="Apples"),
                          ds[elin].sel(Item="Apples"))
    
    assert np.array_equal(result_scalar_item_sub[elout].sel(Item="Apples"),
                          ds[elout].sel(Item="Apples"))

    # Test with array scale on all items
    result_array = fbs.scale_add(elin, elout, array_scale)
    assert np.array_equal(result_array[elin], ds[elin]*array_scale)
    assert np.array_equal(result_array[elout], ds[elout] + (result_array[elin]
                                                            - ds[elin]))

    # Test with DataArray scale on years only
    result_xarray_year = fbs.scale_add(elin, elout, xarray_year_scale)
    assert np.array_equal(result_xarray_year[elin], ds[elin]*xarray_year_scale)
    assert np.array_equal(result_xarray_year[elout], ds[elout]
                          + (result_xarray_year[elin] - ds[elin]))

    # Test with DataArray scale on items only
    result_xarray_item = fbs.scale_add(elin, elout, xarray_item_scale)
    assert np.array_equal(result_xarray_item[elin], ds[elin]*xarray_item_scale)
    assert np.array_equal(result_xarray_item[elout], ds[elout]
                          + (result_xarray_item[elin] - ds[elin]))

def test_scale_element():
    
    items = ["Beef", "Apples"]
    years = [2020, 2021]

    ds = xr.Dataset(
        data_vars=dict(
            imports=(["Year", "Item"], [[10, 20], [30, 40]]),
            exports=(["Year", "Item"], [[5, 10], [15, 20]]),
            production=(["Year", "Item"], [[50, 60], [70, 80]])
            ),
        
        coords=dict(Item=("Item", items), Year=("Year", years))
        )
    
    fbs = FoodBalanceSheet(ds)

    # Scale single element by a single value
    result_single = fbs.scale_element("production", 0.5)
    xr.testing.assert_equal(result_single["production"], 0.5*ds["production"])
    xr.testing.assert_equal(result_single["imports"], ds["imports"])

    # Scale item subset
    result_subset = fbs.scale_element("production", 0.5, items="Beef")
    xr.testing.assert_equal(result_subset["production"],
                            [0.5, 1.0]*ds["production"])

    # Scale element array
    elements = ["production", "imports"]
    result_elements = fbs.scale_element(elements, 0.5)
    xr.testing.assert_equal(result_elements[elements], 0.5*ds[elements])
    xr.testing.assert_equal(result_elements["exports"], ds["exports"])

    # Scale by array of values
    scale_arr = [0.5, 2.0]
    result_arr = fbs.scale_element("production", scale_arr)
    xr.testing.assert_equal(result_arr["production"],
                            [0.5, 2.0]*ds["production"])


    # Scale by array of xarray along one named dimension
    scale_year = xr.DataArray([0.5, 2.0], dims="Year", coords={"Year":years})
    result_year = fbs.scale_element("production", scale_year)
    xr.testing.assert_equal(result_year["production"],
                            [[0.5], [2.0]]*ds["production"])



"""Impact module.
"""

import numpy as np
import xarray as xr
from agrifoodpy.array_accessor import XarrayAccessorBase

def impact(items, regions, quantities, datasets=None, long_format=True):
    """Impact style dataset constructor

    Parameters
    ----------
    items : (ny,) array_like
        The Item identifying ID or strings for which coordinates will be
        created.
    regions : (nr,) array_like, optional
        The region identifying ID or strings for which coordinates will be
        created.
    quantities : (ny, nr) ndarray
        Array containing the quantities for each combination of `Item` and
        `Region`.
    datasets : (nd,) array_like, optional
        Array with model name strings
    long_format : bool
        Boolean flag to interpret data in long or wide format

    Returns
    -------
    data : xarray.Dataset
        Impact dataset containing the impact for each `Item` and
        `Region` with one dataarray per element in `dataset`.
    """

    # if the input has a single element, proceed with long format
    if np.isscalar(quantities):
        long_format = True

    quantities = np.array(quantities)

    # Identify unique values in coordinates
    _items = np.unique(items)
    _regions = np.unique(regions)

    coords = {"Item" : _items,
              "Region" : _regions}

    # find positions in output array to organize data
    ii = [np.searchsorted(_items, items), np.searchsorted(_regions, regions)]
    size = (len(_items), len(_regions))

    # Create empty dataset
    data = xr.Dataset(coords = coords)

    if long_format:
        ndims = 2
    else:
        ndims = 3

    # make sure the long format has two dimensions
    # One along items and regions, one along datasets
    while len(quantities.shape) < ndims:
        quantities = np.expand_dims(quantities, axis=0)

    # If no dataset names are given, then create generic ones, one for each
    # dataset
    if datasets is None:
        datasets = [f"Impact {id}" for id in range(quantities.shape[0])]

    # Else, if a single string is given, transform to list. If doesn't match
    # the number of datasets created above, this will spit a list later.
    elif isinstance(datasets, str):
        datasets = [datasets]

    if long_format:
        # Create a datasets, one at a time
        for id, dataset in enumerate(datasets):
            values = np.zeros(size)*np.nan
            values[tuple(ii)] = quantities[id]
            data[dataset] = (coords, values)

    else:
        quantities = quantities[:, ii[0]]
        quantities = quantities[..., ii[1]]

        # Create a datasets, one at a time
        for id, dataset in enumerate(datasets):
            values = quantities[id]
            data[dataset] = (coords, values)

    return data

@xr.register_dataset_accessor("impact")
class Impact(XarrayAccessorBase):

    def __init__(self, xarray_obj):
        self._validate(xarray_obj)
        self._obj = xarray_obj

    @staticmethod
    def _validate(obj):
        """Check that the Impact object is an xarray.DataArray or xarray.Dataset
        """
        if not isinstance(obj, xr.Dataset):
            raise TypeError("Food Balance Sheet must be an xarray.DataSet")

    def match(self, matching_matrix):
        """Matches an impact dataset to a new item base using a matching matrix

        Parameters
        ----------
        impact: xarray.DataSet
            xarray dataset including a list of items and impacts
        matching_matrix: pandas dataframe
            Defines how items are matched from the input to the output datasets,
            with the values of the matrix indicating the scaling of the
            impact quantities. Column names indicate the original item list, while
            row names indicate the new item list

        Returns
        -------
        dataset_out : xarray.Dataset
            FAOSTAT formatted Food Supply dataset with scaled quantities.

        """

        impact = self._obj

        out_items = matching_matrix["Item Code"]

        in_items = impact.Item.values

        # First column is the item code column
        in_items_mat = matching_matrix.columns[1:]

        assert np.equal(in_items, in_items_mat).all() , \
            "Input items do not match assignment matrix"

        # Again, we avoid first column
        mat = matching_matrix.iloc[:, 1:].fillna(0).to_numpy()

        dataset_out = xr.Dataset(
            coords = dict(
                Item=("Item", out_items),
            )
        )

        for var in list(impact.keys()):
            data_out = np.matmul(mat, impact[var].values)
            dataset_out = dataset_out.assign({var:("Item", data_out)})

        return dataset_out


import numpy as np
import xarray as xr
from agrifoodpy.land.model import land_sequestration

def test_land_sequestration():

    # basic test with one category and one fraction
    lda = xr.DataArray([[0,1],[1,0]], coords={"x":[0,1], "y":[0,1]},
                       name = "land")
    
    result = land_sequestration(lda, 0, 0.5, 10)
    truth = xr.DataArray(data=10)

    xr.testing.assert_equal(result, truth)

    # test with multiple categories and one fraction
    result = land_sequestration(lda, [0,1], 0.5, 10)
    truth = xr.DataArray(data=20)

    xr.testing.assert_equal(result, truth)

    # test with multiple categories and multiple fractions
    result = land_sequestration(lda, [0,1], [0.5, 0.1], 10)
    truth = xr.DataArray(data=12)

    xr.testing.assert_equal(result, truth)

    # test with years value
    result = land_sequestration(lda, 0, 0.5, 10, years=10, growth_timescale=10)
    truth = xr.DataArray(data=np.arange(11), coords={"Year":np.arange(11)})

    xr.testing.assert_allclose(result, truth)

    # test with years array
    result = land_sequestration(lda, 0, 0.5, 10, years=np.arange(1960,2020),
                                growth_timescale=10)
    truth_data = np.concatenate((np.arange(11), np.ones(2020-1960-11)*10))
    truth = xr.DataArray(data=truth_data, coords={"Year":np.arange(1960, 2020)})

    xr.testing.assert_allclose(result, truth)

    # test with logistic growth function
    result = land_sequestration(lda, 0, 0.5, 10, years=10, growth_timescale=10,
                                growth="logistic")
    truth = xr.DataArray([0.06692851, 0.1798621 , 0.47425873, 1.19202922,
                        2.68941421, 5.        , 7.31058579, 8.80797078,
                        9.52574127, 9.8201379 , 10.],
                        coords={"Year":np.arange(11)})

    xr.testing.assert_allclose(result, truth)


"""Land module.
"""

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

@xr.register_dataarray_accessor("land")
class LandDataArray:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._validate(xarray_obj)

    @staticmethod
    def _validate(obj):
        """Validate land xarray, checking it is a DataSet and has the minimum set
        of coordinates
        """
        if not isinstance(obj, xr.DataArray):
            raise TypeError("Land array must be an xarray.Dataarray")
    
        if "x" not in obj.dims or "y" not in obj.dims:
            raise AttributeError("Land array must have 'x' and 'y' dimensions")

    def plot(self, ax=None, **kwargs):
        """Plot a LandDataArray
        
        Generates a plot of a LandDataArray using matplotlib imshow, without
        interpolation and setting the origin low to align north at the top.

        Parameters
        ----------
        ax : matplotlib.pyplot.Artist
            Axes on which to draw the plot
        **kwargs : dict
            Style options to be passed to the imshow function.
        
        Returns
        -------
        ax : matplotlib axes instance
        """

        map = self._obj

        if ax is None:
            f, ax = plt.subplots()

        # Get plot ranges

        dx_low = (map.x.values[1] - map.x.values[0])/2
        dx_high = (map.x.values[-1] - map.x.values[-2])/2
        dy_low = (map.y.values[1] - map.y.values[0])/2
        dy_high = (map.y.values[-1] - map.y.values[-2])/2

        xmin, xmax = map.x.values[[0, -1]]
        ymin, ymax = map.y.values[[0, -1]]

        ax.imshow(map, interpolation="none", origin="lower",
                  extent=[xmin-dx_low, xmax+dx_high, ymin-dy_low, ymax+dy_high])
        
        return ax
        
    def area_by_type(self, values = None, dim = None):
        """Area per map category in a LandDataArray
        
        Returns a DataArray with the total number of pixels for each category or
        category subset of the LandDataArray. 

        Parameters
        ----------
        values : int, array
            List of category types to return the total area for. If not set, the
            function returns areas for all values found on the map, excluding
            nan values.
        dim : string
            Name to assign to the categories coordinate. If not set, the input
            DataArray name is used instead.
        
        Returns
        -------
        xarray.DataArray
            Array with the corresponding areas overlaps for each category
            combination.
        """

        map = self._obj
        ones = xr.ones_like(map)

        if dim is None:
            dim = map.name

        if values is None:
            values = np.unique(map)
        else:
            values = np.array(values)

        # Prevent nan values from being counted
        nan_indices = np.isnan(values)
        values = values[~nan_indices]
        area = [ones.where(map==value).sum() for value in values]
        
        area_arr = xr.DataArray(area, dims=dim, coords={dim:values})

        return area_arr

    def area_overlap(self, map_right, values_left = None, values_right = None, dim_left=None, dim_right=None):
        """Area overlap of selected categories between two maps
        
        Returns a DataArray with the total number of pixels for each combination
        of categories from the left and right map selected categories. Casa

        Parameters
        ----------
        map_right : xarray.DataArray
            LandDataArray style DataArray to compare overlapping areas with
        values_left
            List of category types from the left map to return the total area
            overlaps for.
            If not set, all category types are used, except nan values.
        values_right : int, array
            List of category types from the right map to return the total area
            overlaps for.
            If not set, all category types are used, except nan values.
        dim_left : string
            Names to assign to the category coordinates on the output DataArray.
            If not set, the input DataArray name is used instead.
        dim_right : string
            Names to assign to the category coordinates on the output DataArray.
            If not set, the input DataArray name is used instead.
        
        Returns
        -------
        area_arr : xarray.DataArray
            Array with the corresponding areas for each category type

        """
        map_left = self._obj

        # Check that both maps have the same dimensions and coordinates. if not,
        # this raises a ValueError (alternatively, align the maps and use )
        xr.align(map_left, map_right, join='exact')

        if dim_left is None:
            dim_left = map_left.name

        if dim_right is None:
            dim_right = map_right.name

        if values_left is None:
            values_left = np.unique(map_left)
        else:
            values_left = np.array(values_left)

        if values_right is None:
            values_right = np.unique(map_right)
        else:
            values_right = np.array(values_right)

        # Prevent nan values from being counted
        nan_indices_left = np.isnan(values_left)
        nan_indices_right = np.isnan(values_right)

        values_left = values_left[~nan_indices_left]
        values_right = values_right[~nan_indices_right]

        ones = xr.ones_like(map_left)
        area = [[ones.where(map_left==vl).where(map_right==vr).sum().values for vr in values_right] for vl in values_left]

        area_arr = xr.DataArray(area, dims=[dim_left, dim_right], coords={dim_left:values_left, dim_right:values_right})

        return area_arr
    
    def category_match(self, map_right, values_left=None, values_right=None,
                       join="left"):
        """Returns a land Dataarray with values where a selected overlap occurs
        between categories from two maps. This returns the values from the left
        map where coincidence occurs between the left and right map.

        Parameters
        ----------
        map_right : xarray.DataArray
            LandDataArray style DataArray to compare overlapping areas with
        values_left : int, array
            List of category types from the left map to match.
            If not set, all category types are used, except nan values.
        values_right : int, array
            List of category types from the right map to match.
            If not set, all category types are used, except nan values.
                
        Returns
        -------
        category_match : xarray.DataArray
            Land DataArray with values from the left map where overlap occurs.
            All other positions are assign a nan value.
        """
        map_left = self._obj

        # Check input values
        if values_left is None:
            values_left = np.unique(map_left)
        else:
            values_left = np.array(values_left)

        if values_right is None:
            values_right = np.unique(map_right)
        else:
            values_right = np.array(values_right)

        # Align maps to the left so they have the same dimensions
        map_left, map_right = xr.align(map_left, map_right, join=join)

        shape = map_left.shape

        left_match = np.in1d(map_left, values_left).reshape(shape)
        right_match = np.in1d(map_right, values_right).reshape(shape)

        category_match = map_left.where(left_match).where(right_match)

        return category_match



import numpy as np
import xarray as xr
from agrifoodpy.array_accessor import XarrayAccessorBase

def test_add_years():

    years = np.arange(2010, 2013)
    data = np.random.rand(3, 2, 2)
    new_years = [2013, 2014, 2015]
    expected_years = np.concatenate([years, new_years])

    ds = xr.Dataset({"data": (("Year", "X", "Y"), data)},
                      coords={"Year": years, "X": [0, 1], "Y": [0, 1]})
    
    fbs = XarrayAccessorBase(ds)

    # Test adding single year

    result_single = fbs.add_years(new_years[0])
    expected_years_single = np.concatenate([years, [new_years[0]]])

    assert np.array_equal(result_single["Year"], expected_years_single)
    assert np.isnan(result_single["data"].loc[{"Year":new_years[0]}].to_numpy()
                    ).all()

    # Test adding years with "empty" projection
    result_empty = fbs.add_years(new_years, projection="empty")

    assert np.array_equal(result_empty["Year"], expected_years)
    assert np.isnan(result_empty["data"].loc[{"Year":new_years}].to_numpy()
                    ).all()

    # Test adding years with "constant" projection
    result_constant = fbs.add_years(new_years, projection="constant")
    last_year_data = ds["data"].isel(Year=-1).values

    assert np.array_equal(result_constant["Year"], expected_years)
    for year in new_years:
        assert np.array_equal(result_constant["data"].loc[{"Year":year}].values,
                              last_year_data)

    # Test adding years with specific projection
    proj = [0.5, 0.6, 0.7]  # Scaling factors
    result_projection = fbs.add_years(new_years, projection=proj)
    last_year_data = result_projection["data"].loc[dict(Year=years[-1])].values
    expected_data = [last_year_data * proj[i] for i in range(len(proj))]

    assert np.array_equal(result_projection["Year"], expected_years)
    assert np.allclose(result_projection["data"].loc[{"Year":new_years}].values,
                       expected_data)

    # Test for duplicate years
    new_years_duplicate = [2013, 2013, 2014, 2015]
    result_duplicate = fbs.add_years(new_years_duplicate)

    assert np.array_equal(result_duplicate["Year"], expected_years)

def test_add_items():

    items = ["Beef", "Apples", "Poultry"]
    item_origin = ["Animal", "Vegetal", "Animal"]
    new_items = ["Tomatoes", "Potatoes", "Eggs"]

    data = np.random.rand(3, 2, 2)
    expected_items = np.concatenate([items, new_items])

    ds = xr.Dataset({"data": (("Item", "X", "Y"), data)},
                    coords={"Item": items, "X": [0, 1], "Y": [0, 1]})
    ds = ds.assign_coords({"Item_origin":("Item", item_origin)})

    fbs = XarrayAccessorBase(ds)

    # Test adding new items
    result_add = fbs.add_items(new_items)

    assert np.array_equal(result_add["Item"].values, expected_items)
    assert np.isnan(result_add["data"].sel(Item=new_items).values).all()

    # Test adding new items copying from single existing one
    result_copy = fbs.add_items(new_items, copy_from="Beef")

    assert np.array_equal(result_copy["Item"], expected_items)
    for item_i in new_items:
        assert np.array_equal(result_copy["data"].sel(Item=item_i),
                              ds.data.sel(Item="Beef"))

    # Test adding new items copying from existing array
    result_copy_multiple = fbs.add_items(new_items, copy_from=["Beef",
                                                               "Apples",
                                                               "Poultry"])

    assert np.array_equal(result_copy_multiple["Item"], expected_items)
    assert np.array_equal(result_copy_multiple["data"].sel(Item=new_items),
                          ds.data.sel(Item=["Beef", "Apples", "Poultry"]))
  
def test_add_regions():

    regions = [1, 2, 3]
    region_name = ["UK", "US", "Chile"]
    new_regions = [4, 5, 6]

    data = np.random.rand(3, 2, 2)
    expected_regions = np.concatenate([regions, new_regions])

    ds = xr.Dataset({"data": (("Region", "X", "Y"), data)},
                    coords={"Region": regions, "X": [0, 1], "Y": [0, 1]})
    ds = ds.assign_coords({"Region_name":("Region", region_name)})

    fbs = XarrayAccessorBase(ds)

    # Test adding new regions
    result_add = fbs.add_regions(new_regions)

    assert np.array_equal(result_add["Region"], expected_regions)
    assert np.isnan(result_add["data"].sel(Region=new_regions).values).all()

    # Test adding new regions copying from single existing one
    result_copy = fbs.add_regions(new_regions, copy_from=1)

    assert np.array_equal(result_copy["Region"], expected_regions)
    for region_i in new_regions:
        assert np.array_equal(result_copy["data"].sel(Region=region_i),
                              ds.data.sel(Region=1))

    # Test adding new regions copying from existing array
    result_copy_multiple = fbs.add_regions(new_regions, copy_from=[1, 2, 3])

    assert np.array_equal(result_copy_multiple["Region"], expected_regions)
    assert np.array_equal(result_copy_multiple["data"].sel(Region=new_regions),
                          ds.data.sel(Region=[1, 2, 3]))
    
def test_group_sum():
    items = ["Beef", "Apples", "Poultry"]
    years = np.arange(1990, 1992)
    regions = ["UK", "US"]
    item_origin = ["Animal", "Vegetal", "Animal"]

    data = np.arange(12).reshape(3, 2, 2)

    ds = xr.Dataset({"data": (("Item", "Year", "Region"), data)},
                    coords={"Item": items,
                            "Year": years,
                            "Region": regions})
    # Label dimension
    ds = ds.assign_coords({"Item_origin":("Item", item_origin)})

    fbs = XarrayAccessorBase(ds)

    # Basic test on non-dimension coordinate
    sum_dim = "Item_origin"
    result = fbs.group_sum(sum_dim)

    sum_dim_values = np.unique(item_origin)
    truth_data = [data[0]+data[2], data[1]]
    truth = xr.Dataset({"data": ((sum_dim, "Year", "Region"), truth_data)},
                    coords={"Year": years,
                            "Region": regions,
                            sum_dim: sum_dim_values})
    
    assert result.equals(truth)
    
    # Basic test on non-dimension coordinate, renaming output dimension
    sum_dim = "Item_origin"
    out_dim = "Elements"
    result = fbs.group_sum(sum_dim, new_name=out_dim)

    sum_dim_values = np.unique(item_origin)
    truth_data = [data[0]+data[2], data[1]]
    truth = xr.Dataset({"data": ((out_dim, "Year", "Region"), truth_data)},
                    coords={"Year": years,
                            "Region": regions,
                            out_dim: sum_dim_values})

    assert result.equals(truth)

    # Test on named dimension
    sum_dim = "Item"
    result = fbs.group_sum(sum_dim)

    truth = xr.Dataset({"data": (("Item", "Year", "Region"), data)},
                    coords={"Year": years,
                            "Region": regions,
                            "Item": items}).sel(Item=np.unique(items))

    assert result.equals(truth)
    


""" Module for land intervention models

"""

import numpy as np
from agrifoodpy.land.land import LandDataArray

def land_sequestration(land_da, use_id, fraction, max_seq, years=None,
                       growth_timescale=10, growth="linear", ha_per_pixel=1):
    
    """Additional land use sequestration model.
    
    Computes the anual additional sequestration from land use change as a
    function of the different land category converted fractional areas.
    
    Given a Land Data Array map with pixel id values for the different land use
    types, the model computes additional sequestration from land given the new
    value in [t CO2e / yr].

    Parameters
    ----------
    land : xarray.Dataarray
        Input land array containing the id id value with the dominant land use
        type on that pixel.
    use_id : int, array
        Land category identifiers for the land uses to be converted.
    fraction : float, array
        Fraction of each repurposed land category  
    max_seq : float
        Maximum sequestration achieved at the end of the growth period in
        [t CO2e / yr]
    growth : str
        Type of sequestration growth. Can be one of "linear" and "logistic"
    growth_timescale : float
        Time in years for land to reach full sequestration potential max_seq
    years : int, array
        Year range length, or array of years for which the sequestration is
        computed. If not set, stationary maximum values are returned
    ha_per_pixel : float
        Area per pixel in hectares.

    Returns
    -------
    seq : xarray.DataArray
        DataArray with the per year sequestration
    """
    
    if np.isscalar(use_id):
        use_id = np.array(use_id)

    # If single scalar value, use the same for all categories
    if np.isscalar(fraction):
        fraction = np.ones_like(use_id)*fraction
    else:
        fraction = np.array(fraction)

    if not (fraction >= 0).all() and (fraction <= 1).all():
        raise ValueError("Input fraction values must be between 0 and 1")
            
    pixel_count_category = land_da.land.area_by_type(values=use_id)
    
    # area in hectares
    area_category = pixel_count_category * ha_per_pixel

    if growth == "linear":
        from agrifoodpy.utils.scaling import linear_scale as growth_shape
    elif growth == "logistic":
        from agrifoodpy.utils.scaling import logistic_scale as growth_shape
    else:
        raise ValueError("Growth must be one of 'linear' or 'logistic'")
    
    if years is not None:
        # single scalar value
        if np.isscalar(years):
            scale = growth_shape(0, 0, growth_timescale, years, 0, 1)
        # year range
        else:
            y0 = np.max((0, np.min(years)))
            y1 = np.max((0, np.min(years)))
            y2 = growth_timescale + y1
            y3 = np.max(years)
            scale = growth_shape(y0, y1, y2, y3, 0, 1)
    else:
        scale = 1

    # agroforestry    
    area = area_category * fraction
    total_seq = area.sum() * scale * max_seq

    return total_seq


""" Module for impact intervention models
"""

import xarray as xr
import numpy as np

def fbs_impacts(fbs, impact_element, population=None, sum_dims=None):
    """Computes total impacts from quantities in a food balance sheet Dataset or
    food element quantity DataArray, summing over items, regions or years if
    instructed.

    Parameters
    ----------
    fbs : xarray.DataArray or xarray.Dataset
        Food Balance Sheet array with food quantities for a set of items,
        regions and/or years.         
    impact_element : xarray.DataArray
        Impact DataArray containing the impacts for set of items, regions and/or
        years.
    population : xarray.DataArray
        If given, the input impacts are considered per-capita values and
        multiplied by the population array

    sum_dims : str
        Dimension labels to sum over
    
    Returns
    -------
    total_impact : xarray.DataArray or xarray.Dataset
        Total impact computed from food balance sheet data and impact array
    """

    total_impact = fbs * impact_element
    if population is not None:
        total_impact *= population

    if sum_dims is not None:
        total_impact = total_impact.sum(dim=sum_dims)
        
    return total_impact

def fair_co2_only(emissions):
    """Simple Interface to FaIR, the Finite-amplitude Impulse-Response
    atmosferic model.

    Computes the concentration, radiative forcing and temperature anomaly for an 
    array of CO2 emissions per year assuming a clean atmosphere and default
    values for amosferic parameters.

    Parameters
    ----------
    emissions : xarray.DataArray or xarray.Dataset
        Array containing GHG emissions in Gt CO2e per year 

    Returns
    -------
    T : xarray.DataArray
        Temperature anomaly in Kelvin degrees at the zero layer
    C : xarray.DataArray
        Atmosferic CO2e concetration in ppm 
    F : xarray.DataArray
        Effective radiative forcing in W m^-2
    """

    from fair import FAIR
    from fair.interface import fill, initialise
    f = FAIR()

    years = np.unique(emissions.Year.values)

    # Configure method, timebounds, and labels
    f.ghg_method='myhre1998'
    f.define_time(years[0]-0.5, years[-1]+0.5, 1)
    f.define_scenarios(["default"])
    f.define_configs(["default"])

    # Define CO2 as the only specie
    species = ['CO2']
    properties = {
        'CO2': {
            'type': 'co2',
            'input_mode': 'emissions',
            # it doesn't behave as a GHG itself in the model, but as a precursor
            'greenhouse_gas': True,  
            'aerosol_chemistry_from_emissions': False,
            'aerosol_chemistry_from_concentration': False,
        }}
    
    f.define_species(species, properties)

    # Allocate arrays
    f.allocate()

    # Set default values
    fill(f.climate_configs["ocean_heat_transfer"], [1.1, 1.6, 0.9],
         config='default')
    
    fill(f.climate_configs["ocean_heat_capacity"], [8, 14, 100],
         config='default')
    
    fill(f.climate_configs["deep_ocean_efficacy"], 1.1, config='default')

    # Set initial conditions.
    initialise(f.concentration, 278.3, specie='CO2')
    initialise(f.forcing, 0)
    initialise(f.temperature, 0)
    initialise(f.cumulative_emissions, 0)
    initialise(f.airborne_emissions, 0)

    # Fill species configs
    f.fill_species_configs()
    
    f.emissions.loc[{"scenario":"default",
                     "specie":"CO2",
                     "config":"default"}] = emissions
    
    # Run and return
    f.run(progress=False)

    return_dict = {"scenario":"default", "config":"default"}

    T = f.temperature.sel(return_dict).drop_vars(
        ["scenario", "config", "layer"]).squeeze()
    
    C = f.concentration.sel(return_dict).drop_vars(
        ["scenario", "config", "specie"]).squeeze()
    
    F = f.forcing.sel(return_dict).drop_vars(
        ["scenario", "config", "specie"]).squeeze()

    return T.sel(layer=0), C, F



import xarray as xr
import numpy as np

class XarrayAccessorBase(object):
    """Common logic for for both data structures (DataArray and Dataset).

    http://xarray.pydata.org/en/stable/internals.html#extending-xarray
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def add_items(self, items, copy_from=None):
        """Extends the item list of an input xarray object according to the
        defined input item list

        Parameters
        ----------
        items : list, int, string
            list of item names to be added to the data
        copy_from : list, int, string, optional
            If provided, this is the list of items already on the array
            to copy data from.
        
        Returns
        -------
        out : xarray.Dataset, xarray.DataArray
            Xarray object with new items added.
        """

        fbs = self._obj

        if np.isscalar(items):
            items = [items]
        
        # Check for duplicates
        indexes = np.unique(items, return_index=True)[1]
        items = [items[index] for index in sorted(indexes)] #issues with np.unique

        new_items = xr.DataArray(data = np.ones(len(items)),
                                coords = {"Item":items})

        if copy_from is not None:
            fbs_pivot = fbs.sel(Item=copy_from)
            fbs_pivot["Item"] = items
            new_fbs = fbs_pivot*new_items
            
        else:
            new_items *= np.nan
            new_fbs = fbs.isel(Item=0)*new_items
            
        # Concatenate to input array
        out = xr.concat([fbs, new_fbs], dim="Item") 
            
        return out

    def add_regions(self, regions, copy_from=None):
        """Extends the region list of an input xarray object according to the
        defined input region list

        Parameters
        ----------
        regions : list, int, string
            list of region names to be added to the data
        copy_from : list, int, string
            If provided, this is the list of regions already on the object to
            copy data from.
        labels : dict 
            Dictionary containing the new label for the regions matched to its
            corresponding label coordinate.
        
        Returns
        -------
        out : xarray.Dataset, xarray.DataArray
            Xarray object with new regions added.
        """
        
        fbs = self._obj

        if np.isscalar(regions):
            regions = [regions]

        indexes = np.unique(regions, return_index=True)[1]
        regions = [regions[index] for index in sorted(indexes)]
        
        new_regions = xr.DataArray(data = np.ones(len(regions)),
                                coords = {"Region":regions})

        if copy_from is not None:
            fbs_pivot = fbs.sel(Region=copy_from)
            fbs_pivot["Region"] = regions
            new_fbs = fbs_pivot*new_regions
            
        else:
            new_regions *= np.nan
            new_fbs = fbs.isel(Region=0)*new_regions
            
        # Concatenate to input array
        out = xr.concat([fbs, new_fbs], dim="Region") 
            
        return out
    
    def add_years(self, years, projection="empty"):
        """Extends the year range of an xarray object according to the defined
        maximum year

        Parameters
        ----------
        years : list, int
            list of years to be added to the data
        projection : string or array_like
            Projection mode. If "constant", the last year of the input array
            is copied to every new year. If "empty", values are initialized and
            set to zero. If a float array is given, these are used to populate
            the new year using a scaling of the last year of the array

        Returns
        -------
        out : xarray.Dataset, xarray.DataArray
            Xarray object with new years added.
        """

        fbs = self._obj

        if np.isscalar(years):
            years = [years]

        indexes = np.unique(years, return_index=True)[1]
        years = [years[index] for index in sorted(indexes)]
        
        if isinstance(projection, str):
            if projection == "empty":
                data = np.zeros(len(years))*np.nan
            elif projection == "constant":
                data = np.ones(len(years))
            else:
                raise ValueError("If a string, 'projection' can only be 'empty'\
                                 or 'constant'")
            
        else:
            # Should raise an exception if sizes do not match
            data = np.ones(len(years)) * projection
        
        
        new_years = xr.DataArray(data=data,
                                    coords = {"Year":years})

        # Select last year as pivot
        fbs_pivot = fbs.isel(Year=-1)
        
        # Create DS or DA by multiplying along last value
        new_fbs = fbs_pivot*new_years
        
        # Concatenate to input array
        out = xr.concat([fbs, new_fbs], dim="Year") 
            
        return out

    def group_sum(self, coordinate, new_name=None):
        """Sums quantities over items of a equal group labels and, optionally,
        renames the groups label coordinate.

        Parameters
        ----------
        fbs : xarray.Dataset
            Input xarray object
        coordinate : string
            Coordinate name to group elements and sum over 
        new_name : string, optional 
            New name for the collapsed coordinate
        

        Returns
        -------
        fbs : xarray.Dataset, xarray.DataArray
            Xarray object with new coordinate base.
        """
        
        fbs = self._obj

        grouped_fbs = fbs.groupby(coordinate).sum()

        if new_name is not None:
            grouped_fbs = grouped_fbs.rename({coordinate:new_name})

        return grouped_fbs




"""Scaling utilities module"""

import numpy as np
import xarray as xr

def logistic_scale(y0, y1, y2, y3, c_init, c_end):
    """
    Create an xarray DataArray with a logistic growth interval

    Parameters:
        y0 : (int)
            The initial year.
        y1 : (int)
            The year when the values start to transition.
        y2 : (int)
            The year when the transition completes.
        y3 : (int)
            The final year.
        c_init : (float)
            The initial constant value.
        c_end : (float)
            The final constant value.
        
    Returns: xarray DataArray
        An xarray DataArray object with 'year' as the coordinate and values set
        by a logistic growth between the user defined intervals.
    """

    # Create arrays and set values between y0 and y1 to c_init
    years = np.arange(y0, y3+1)
    values = np.ones_like(years, dtype=float) * c_init

    # Set values between y1 and y2 using a logistic curve
    var_segment = np.logical_and(years >= y1, years < y2)
    t = (years[var_segment] - y1) / (y2 - y1)
    values[var_segment] = c_init + \
                          (c_end - c_init) *(1 / (1 + np.exp(-10 * (t - 0.5))))

    # Set values between y2 and y3 to c_end
    values[years >= y2] = c_end

    data_array = xr.DataArray(values, dims='Year', coords={'Year': years})
    return data_array

def linear_scale(y0, y1, y2, y3, c_init, c_end):
    """
    Create an xarray DataArray with a single coordinate called 'year'.
    
    The values from the first year 'y0' up to a given year 'y1' will be
    constant, then they will vary linearly between 'y1' and another given year
    'y2', and from 'y2' until the end of the array 'y3', the values will
    continue with a constant value.
    
    Parameters:
        y0 : (int)
            Starting year.
        y1 : (int)
            Year where the linear variation starts.
        y2 : (int)
            Year where the linear variation ends.
        y3 : (int)
            Last year in the array.
        c_init : (float)
            Value to use for initial constant scale segment.
        c_end : (float)
    
    Returns: xr.DataArray
        An xarray DataArray object with 'year' as the coordinate and values set
        by a linear growth between the user defined intervals.
    """
    
    # Create arrays and set values between y0 and y1 to c_init
    years = np.arange(y0, y3 + 1)
    values = np.ones_like(years, dtype=float) * c_init

    # Set values between y1 and y2 using straight line.
    var_segment = np.logical_and(years >= y1, years < y2)
    if y2 == y1:
        slope = c_end - c_init
    else:
        slope = float((c_end - c_init) / (y2 - y1))
    values[var_segment] = slope * (years[var_segment] - y1) + c_init
    
    # Set values between y2 and y3 to c_end
    values[years >= y2] = c_end
        
    data_array = xr.DataArray(values, coords={'Year': years}, dims=['Year'])
    
    return data_array


"""Population module.
"""

import numpy as np
import xarray as xr

from agrifoodpy.array_accessor import XarrayAccessorBase

def population(years, regions, quantities, datasets=None, long_format=True):
    """Population style dataset constructor

    Parameters
    ----------
    years : (ny,) array_like
        The year values for which coordinates will be created.
    regions : (nr,) array_like, optional
        The region identifying ID or strings for which coordinates will be
        created.
    quantities : ([nd], ny, nr) ndarray
        Array containing the population for each combination of `Year` and
        `Region`.
    datasets : (nd,) array_like, optional
        Array with model name strings. If provided, a dataset is created
        for each element in `datasets` with the quantities being each of the
        sub-arrays indexed by the last coordinate of the input array.
    long_format : bool
        Boolean flag to interpret data in long or wide format

    Returns
    -------
    data : xarray.Dataset
        Population dataset containing the population for each `Year` and
        `Region` with one dataarray per element in `dataset`.
    """

    # if the input has a single element, proceed with long format
    if np.isscalar(quantities):
        long_format = True

    quantities = np.array(quantities)

    # Identify unique values in coordinates
    _years = np.unique(years)
    _regions = np.unique(regions)
    coords = {"Year" : _years,
              "Region" : _regions}

    # find positions in output array to organize data
    ii = [np.searchsorted(_years, years), np.searchsorted(_regions, regions)]
    size = (len(_years), len(_regions))

    # Create empty dataset
    data = xr.Dataset(coords = coords)

    if long_format:
        # dataset, quantities
        ndims = 2
    else:
        # dataset, year, region
        ndims = 3

    # make sure the long format has the right number of dimensions
    while len(quantities.shape) < ndims:
        quantities = np.expand_dims(quantities, axis=0)

    # If no dataset names are given, then create generic ones, one for each
    # dataset
    if datasets is None:
        datasets = [f"Population {id}" for id in range(quantities.shape[0])]

    # Else, if a single string is given, transform to list. If doesn't match
    # the number of datasets created above, xarray will return an error.
    elif isinstance(datasets, str):
        datasets = [datasets]

    if long_format:
        # Create a datasets, one at a time
        for id, dataset in enumerate(datasets):
            values = np.zeros(size)*np.nan
            values[tuple(ii)] = quantities[id]
            data[dataset] = (coords, values)

    else:
        quantities = quantities[:, ii[0]]
        if len(quantities.shape) < ndims:
            quantities = np.expand_dims(quantities, axis=0)

        quantities = quantities[..., ii[1]]
        if len(quantities.shape) < ndims:
            quantities = np.expand_dims(quantities, axis=0)
        # Create a datasets, one at a time
        for id, dataset in enumerate(datasets):
            values = quantities[id]
            data[dataset] = (coords, values)

    return data

@xr.register_dataarray_accessor("pop")
class PopulationDataArray(XarrayAccessorBase):
    pass
